@@ -55,7 +55,7 @@ def gradient_penalty(
     output,
     weight = 10
 ):
-    batch_size = images.shape[0]
+    images.shape[0]
     gradients, *_ = torch_grad(
         outputs = output,
         inputs = images,
@@ -260,17 +260,19 @@ class SelfAttention(nn.Module):
         dim,
         dim_head = 64,
         heads = 8,
-        mask_self_value = -1e2
+        dot_product = False
     ):
         super().__init__()
         self.heads = heads
         self.scale = dim_head ** -0.5
         dim_inner = dim_head * heads
 
-        self.mask_self_value = mask_self_value
+        self.dot_product = dot_product
 
         self.norm = ChannelRMSNorm(dim)
-        self.to_qk = nn.Conv2d(dim, dim_inner, 1, bias = False)
+
+        self.to_q = nn.Conv2d(dim, dim_inner, 1, bias = False)
+        self.to_k = nn.Conv2d(dim, dim_inner, 1, bias = False) if dot_product else None
         self.to_v = nn.Conv2d(dim, dim_inner, 1, bias = False)
 
         self.null_kv = nn.Parameter(torch.randn(2, heads, dim_head))
@@ -289,7 +291,7 @@ class SelfAttention(nn.Module):
         i - source seq (attend from)
         j - target seq (attend to)
         """
-        batch, device = fmap.shape[0], fmap.device
+        batch = fmap.shape[0]
 
         fmap = self.norm(fmap)
 
@@ -297,10 +299,11 @@ class SelfAttention(nn.Module):
 
         h = self.heads
 
-        qk, v = self.to_qk(fmap), self.to_v(fmap)
-        qk, v = map(lambda t: rearrange(t, 'b (h d) x y -> (b h) (x y) d', h = self.heads), (qk, v))
+        q, v = self.to_q(fmap), self.to_v(fmap)
 
-        q, k = qk, qk
+        k = self.to_k(fmap) if exists(self.to_k) else q
+
+        q, k, v = map(lambda t: rearrange(t, 'b (h d) x y -> (b h) (x y) d', h = self.heads), (q, k, v))
 
         # add a null key / value, so network can choose to pay attention to nothing
 
@@ -309,17 +312,16 @@ class SelfAttention(nn.Module):
         k = torch.cat((nk, k), dim = -2)
         v = torch.cat((nv, v), dim = -2)
 
-        # l2 distance
+        # l2 distance or dot product
 
-        sim = -torch.cdist(q, k, p = 2) * self.scale
+        if self.dot_product:
+            sim = einsum('b i d, b j d -> b i j', q, k)
+        else:
+            sim = -(torch.cdist(q, k, p = 2) ** 2)
 
-        # following what was done in reformer for shared query / key space
-        # omit attention to self
+        # scale
 
-        self_mask = torch.eye(sim.shape[-2], device = device, dtype = torch.bool)
-        self_mask = F.pad(self_mask, (1, 0), value = False)
-
-        sim = sim.masked_fill(self_mask, self.mask_self_value)
+        sim = sim * self.scale
 
         # attention
 
@@ -378,7 +380,7 @@ class CrossAttention(nn.Module):
 
         q = rearrange(q, 'b (h d) x y -> (b h) (x y) d', h = self.heads)
 
-        sim = -torch.cdist(q, k, p = 2) * self.scale # l2 distance
+        sim = einsum('b i d, b j d -> b i j', q, k) * self.scale
 
         if exists(mask):
             mask = repeat(mask, 'b j -> (b h) 1 j', h = self.heads)
@@ -501,10 +503,11 @@ class SelfAttentionBlock(nn.Module):
         dim,
         dim_head = 64,
         heads = 8,
-        ff_mult = 4
+        ff_mult = 4,
+        dot_product = False
     ):
         super().__init__()
-        self.attn = SelfAttention(dim = dim, dim_head = dim_head, heads = heads)
+        self.attn = SelfAttention(dim = dim, dim_head = dim_head, heads = heads, dot_product = dot_product)
         self.ff = FeedForward(dim = dim, mult = ff_mult, channel_first = True)
 
     def forward(self, x):
@@ -558,8 +561,8 @@ class Transformer(nn.Module):
 
 # text encoder
 
-@beartype
 class TextEncoder(nn.Module):
+    @beartype
     def __init__(
         self,
         *,
@@ -587,6 +590,7 @@ class TextEncoder(nn.Module):
             heads = heads
         )
 
+    @beartype
     def forward(
         self,
         texts: List[str]
@@ -655,8 +659,8 @@ class StyleNetwork(nn.Module):
 
 # generator
 
-@beartype
 class Generator(nn.Module):
+    @beartype
     def __init__(
         self,
         *,
@@ -671,6 +675,7 @@ class Generator(nn.Module):
         self_attn_resolutions: Tuple[int, ...] = (32, 16),
         self_attn_dim_head = 64,
         self_attn_heads = 8,
+        self_attn_dot_product = True,
         self_ff_mult = 4,
         cross_attn_resolutions: Tuple[int, ...] = (32, 16),
         cross_attn_dim_head = 64,
@@ -689,8 +694,10 @@ class Generator(nn.Module):
         self.text_encoder = text_encoder
 
         self.unconditional = unconditional
+
         assert not (unconditional and exists(text_encoder))
         assert not (unconditional and exists(style_network) and style_network.dim_text_latent > 0)
+        assert unconditional or (exists(text_encoder) and text_encoder.dim == style_network.dim_text_latent), 'the `dim_text_latent` on your StyleNetwork must be equal to the `dim` set for the TextEncoder'
 
         assert is_power_of_two(image_size)
         num_layers = int(log2(image_size) - 1)
@@ -731,7 +738,7 @@ class Generator(nn.Module):
         dim_layers = F.pad(dim_layers, (1, 0), value = dim_latent)
 
         dim_layers = dim_layers.tolist()
-        dim_last = dim_layers[-1]
+        dim_layers[-1]
         dim_pairs = list(zip(dim_layers[:-1], dim_layers[1:]))
 
         self.num_skip_layers_excite = num_skip_layers_excite
@@ -741,7 +748,6 @@ class Generator(nn.Module):
         # go through layers and construct all parameters
 
         for ind, ((dim_in, dim_out), resolution) in enumerate(zip(dim_pairs, resolutions)):
-            is_first = ind == 0
             is_last = (ind + 1) == len(dim_pairs)
             should_upsample = not is_last
             should_skip_layer_excite = num_skip_layers_excite > 0 and (ind + num_skip_layers_excite) < len(dim_pairs)
@@ -773,7 +779,7 @@ class Generator(nn.Module):
                 rgb_upsample = Upsample(channels)
 
             if has_self_attn:
-                self_attn = SelfAttentionBlock(dim_out)
+                self_attn = SelfAttentionBlock(dim_out, dot_product = self_attn_dot_product)
 
             if has_cross_attn:
                 cross_attn = CrossAttentionBlock(dim_out, dim_context = text_encoder.dim)
@@ -805,16 +811,18 @@ class Generator(nn.Module):
     @property
     def device(self):
         return next(self.parameters()).device
-    
+
+    @beartype
     def forward(
         self,
-        noise = None,
         styles = None,
+        noise = None,
         texts: Optional[List[str]] = None,
         global_text_tokens = None,
         fine_text_tokens = None,
         text_mask = None,
-        batch_size = 1
+        batch_size = 1,
+        return_all_rgbs = False
     ):
         # take care of text encodings
         # which requires global text tokens to adaptively select the kernels from the main contribution in the paper
@@ -855,6 +863,10 @@ class Generator(nn.Module):
 
         excitations = [None] * self.num_skip_layers_excite
 
+        # all the rgb's of each layer of the generator is to be saved for multi-resolution input discrimination
+
+        rgbs = []
+
         # main network
 
         for squeeze_excite, (resnet_conv1, act1, resnet_conv2, act2), to_rgb_conv, self_attn, cross_attn, upsample, upsample_rgb in self.layers:
@@ -878,7 +890,11 @@ class Generator(nn.Module):
             if exists(cross_attn):
                 x = cross_attn(x, context = fine_text_tokens, mask = text_mask)
 
-            rgb = rgb + to_rgb_conv(x, mod = next(conv_mods), kernel_mod = next(conv_mods))
+            layer_rgb = to_rgb_conv(x, mod = next(conv_mods), kernel_mod = next(conv_mods))
+
+            rgb = rgb + layer_rgb
+
+            rgbs.append(rgb)
 
             if exists(upsample):
                 x = upsample(x)
@@ -886,6 +902,9 @@ class Generator(nn.Module):
             if exists(upsample_rgb):
                 rgb = upsample_rgb(rgb)
 
+        if return_all_rgbs:
+            return rgb, rgbs
+
         return rgb
 
 # discriminator
@@ -964,6 +983,7 @@ class VisionAidedDiscriminator(nn.Module):
         dim_head = 64,
         heads = 8,
         layer_indices = (-1, -2, -3),
+        conv_dim = None,
         text_dim = None,
         unconditional = False,
         num_conv_kernels = 2
@@ -974,6 +994,7 @@ class VisionAidedDiscriminator(nn.Module):
 
         self.unconditional = unconditional
         text_dim = default(text_dim, dim)
+        conv_dim = default(conv_dim, dim)
 
         self.layer_discriminators = nn.ModuleList([])
         self.layer_indices = layer_indices
@@ -982,12 +1003,12 @@ class VisionAidedDiscriminator(nn.Module):
 
         for _ in layer_indices:
             self.layer_discriminators.append(nn.ModuleList([
-                RandomFixedProjection(dim, dim),
-                conv_klass(dim, dim),
-                nn.Linear(text_dim, dim) if not unconditional else None,
+                RandomFixedProjection(dim, conv_dim),
+                conv_klass(conv_dim, conv_dim),
+                nn.Linear(text_dim, conv_dim) if not unconditional else None,
                 nn.Linear(text_dim, num_conv_kernels) if not unconditional else None,
                 nn.Sequential(
-                    conv2d_3x3(dim, 1),
+                    conv2d_3x3(conv_dim, 1),
                     Rearrange('b 1 ... -> b ...')
                 )
             ]))
@@ -1016,7 +1037,9 @@ class VisionAidedDiscriminator(nn.Module):
             height_width = int(sqrt(rest_tokens.shape[-2])) # assume square
 
             img_fmap = rearrange(rest_tokens, 'b (h w) d -> b d h w', h = height_width)
+
             img_fmap = img_fmap + rearrange(cls_token, 'b 1 d -> b d 1 1 ') # pool the cls token into the rest of the tokens
+            img_fmap = rand_proj(img_fmap)
 
             if self.unconditional:
                 img_fmap = conv(img_fmap)
@@ -1073,8 +1096,8 @@ class Predictor(nn.Module):
         x = x + residual
         return self.to_logits(x)
 
-@beartype
 class Discriminator(nn.Module):
+    @beartype
     def __init__(
         self,
         *,
@@ -1086,6 +1109,7 @@ class Discriminator(nn.Module):
         attn_resolutions: Tuple[int] = (32, 16),
         attn_dim_head = 64,
         attn_heads = 8,
+        self_attn_dot_product = False,
         ff_mult = 4,
         text_encoder: Optional[TextEncoder] = None,
         text_dim = None,
@@ -1165,10 +1189,22 @@ class Discriminator(nn.Module):
                 dim_skip_in, _ = dim_pairs[ind + num_skip_layers_excite]
                 skip_squeeze_excite = SqueezeExcite(dim_in, dim_skip_in)
 
+            # multi-scale rgb input to feature dimension
+
+            from_rgb = None
+            if has_multiscale_input:
+                from_rgb = nn.Conv2d(channels, dim_in, 7, padding = 3)
+
+            # dim in + channels for the main features
+
             dim_in = dim_in + (channels if has_multiscale_input else 0)
 
+            # residual convolution
+
             residual_conv = nn.Conv2d(dim_in, dim_out, 1, stride = (2 if should_downsample else 1))
 
+            # main resnet block
+
             mult = 2 if use_glu else 1
             act_fn = partial(nn.GLU, dim = 1) if use_glu else leaky_relu
 
@@ -1179,6 +1215,8 @@ class Discriminator(nn.Module):
                 act_fn()
             )
 
+            # multi-scale output
+
             multiscale_output_predictor = None
             if has_multiscale_output:
                 multiscale_output_predictor = Predictor(dim_out, num_conv_kernels = num_conv_kernels, unconditional = unconditional)
@@ -1197,9 +1235,10 @@ class Discriminator(nn.Module):
 
             self.layers.append(nn.ModuleList([
                 skip_squeeze_excite,
+                from_rgb,
                 resnet_block,
                 residual_conv,
-                SelfAttentionBlock(dim_out, heads = attn_heads, dim_head = attn_dim_head, ff_mult = ff_mult) if has_attn else None,
+                SelfAttentionBlock(dim_out, heads = attn_heads, dim_head = attn_dim_head, ff_mult = ff_mult, dot_product = self_attn_dot_product) if has_attn else None,
                 multiscale_output_predictor,
                 aux_recon_decoder,
                 Downsample(dim_out) if should_downsample else None,
@@ -1225,11 +1264,17 @@ class Discriminator(nn.Module):
             self.predictor_dims = predictor_dims
             self.text_to_conv_conditioning = nn.Linear(self.text_dim, sum(predictor_dims)) if exists(self.text_dim) else None
 
+    def resize_image_to(self, images, resolution):
+        return F.interpolate(images, resolution, mode = self.resize_mode)
+
+    @beartype
     def forward(
         self,
         images,
+        rgbs: Optional[List[torch.Tensor]] = None,  # multi-resolution inputs (rgbs) from the generator
         texts: Optional[List[str]] = None,
-        text_embeds = None
+        text_embeds = None,
+        real_images = None                          # if this were passed in, the network will automatically append the real to the presumably generated images passed in as the first argument, and generate all intermediate resolutions through resizing and concat appropriately
     ):
         if not self.unconditional:
             assert exists(texts) ^ exists(text_embeds)
@@ -1243,8 +1288,26 @@ class Discriminator(nn.Module):
         else:
             assert not any([*map(exists, (texts, text_embeds))])
 
+        batch = images.shape[0]
+
         x = images
 
+        # if real images are passed in, assume `images` are generated, and take care of all the multi-resolution input. this can also be done externally, in which case `real_images` will not be populated
+
+        has_real_images = exists(real_images)
+        x = torch.cat((x, real_images), dim = 0)
+
+        aux_recon_target = x
+
+        assert not (has_real_images and not exists(rgbs)) 
+
+        if has_real_images:
+            rgbs = [torch.cat((rgb, self.resize_image_to(real_images, rgb.shape[-1])), dim = 0) for rgb in rgbs]
+
+        # index the rgbs by resolution
+
+        rgbs_index = {t.shape[-1]: t for t in rgbs} if exists(rgbs) else {}
+
         # hold multiscale outputs
 
         multiscale_outputs = []
@@ -1257,7 +1320,7 @@ class Discriminator(nn.Module):
 
         excitations = [None] * (self.num_skip_layers_excite + 1) # +1 since first image in pixel space is not excited
 
-        for squeeze_excite, block, residual_fn, attn, predictor, recon_decoder, downsample in self.layers:
+        for squeeze_excite, from_rgb, block, residual_fn, attn, predictor, recon_decoder, downsample in self.layers:
             resolution = x.shape[-1]
 
             if exists(squeeze_excite):
@@ -1266,11 +1329,29 @@ class Discriminator(nn.Module):
 
             excite = safe_unshift(excitations)
             if exists(excite):
+                excite = repeat(excite, 'b ... -> (s b) ...', s = x.shape[0] // excite.shape[0])
                 x = x * excite
 
             if resolution in self.multiscale_input_resolutions:
-                resized_images = F.interpolate(images, resolution, mode = self.resize_mode)
-                x = torch.cat((resized_images, x), dim = 1)
+                images_to_concat = rgbs_index.get(resolution, None)
+
+                # if no rgbs passed in, assume all real images, and just resize, though realistically you would concat fake and real images together using helper function `create_real_fake_rgbs` function
+
+                if not exists(images_to_concat):
+                    images_to_concat = self.resize_image_to(images, resolution)
+
+                images_to_concat = repeat(images_to_concat, 'b ... -> (s b) ...', s = x.shape[0] // images_to_concat.shape[0])
+
+                # concat the rgb (or real images reshaped)
+
+                x = torch.cat((images_to_concat, x), dim = 1)
+
+                # concat the rgb, projected into the feature dimension space
+
+                if exists(from_rgb):
+                    multi_scale_x = from_rgb(images_to_concat)
+                    multi_scale_x = torch.cat((images_to_concat, multi_scale_x), dim = 1)
+                    x = torch.cat((x, multi_scale_x), dim = 0)
 
             residual = residual_fn(x)
             x = block(x)
@@ -1292,17 +1373,19 @@ class Discriminator(nn.Module):
             x = x * self.residual_scale
 
             if exists(recon_decoder):
-                aux_recon_loss = recon_decoder(x, images)
+                aux_recon_target = repeat(aux_recon_target, 'b ... -> (s b) ...', s = x.shape[0] // aux_recon_target.shape[0])
+                aux_recon_loss = recon_decoder(x, aux_recon_target)
                 aux_recon_losses.append(aux_recon_loss)
 
         logits = self.to_logits(x)   
+        logits = rearrange(logits, '(s b) ... -> s b ...', b = batch)
 
         return logits, multiscale_outputs, aux_recon_losses
 
 # gan
 
-@beartype
 class GigaGAN(nn.Module):
+    @beartype
     def __init__(
         self,
         *,
